// Copyright 2020-2024 Yuji Tanaka
// HP : http://galaxystar.image.coocan.jp/
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
/*
 * ESP32_ILI9341_game_BLE_keyboard.ino
 *     Bluetooth Keyboard : "ESP32_BLE_keyboard"
 *
 *   ESP32   : MCU
 *   ILI9341 : LCD panel
 *   XPT2046 : Touch panel
 */ 
 
#ifdef dobogusinclude
#include <spi4teensy3.h>
#endif
#include "BleHidKeyboard.h"

#include <SPI.h>
#include <ILI9341_game.h>
#include <XPT2046_Touchscreen.h>
#include "font16.h"

#define TFT_DC 48
#define TFT_CS 47
#define TFT_RST 45

ILI9341_game tft = ILI9341_game(TFT_CS, TFT_DC, TFT_RST);

#define CS_PIN 39
XPT2046_Touchscreen ts(CS_PIN);
#define TIRQ_PIN 38
#define TS_MINX 400
#define TS_MINY 300
#define TS_MAXX 3900
#define TS_MAXY 3800

#define XMAX 320
#define YMAX 240
#define SIZEX 29
#define SIZEY 32

#define LED   14
 
BleHidKeyboard* ble = 0;

boolean istouched, wastouched;

#define MAX_KEY 63

struct f_key_code {
  int code;
  int x;
  int y;
  int sx;
  int sy;
  char name[8];
  char shift[8];
} key_code[MAX_KEY] = {
  0x1E, 0, 0, 1, 1, "1", "!",
  0x1F, 1, 0, 1, 1, "2", "\"",
  0x20, 2, 0, 1, 1, "3", "#",
  0x21, 3, 0, 1, 1, "4", "$",
  0x22, 4, 0, 1, 1, "5", "%",
  0x23, 5, 0, 1, 1, "6", "&",
  0x24, 6, 0, 1, 1, "7", "'",
  0x25, 7, 0, 1, 1, "8", "(",
  0x26, 8, 0, 1, 1, "9", ")",
  0x27, 9, 0, 1, 1, "0", " ",
  0x2D,10, 0, 1, 1, "-", "=",
  
  0x14, 0, 1, 1, 1, "q", "Q",
  0x1A, 1, 1, 1, 1, "w", "W",
  0x08, 2, 1, 1, 1, "e", "E",
  0x15, 3, 1, 1, 1, "r", "R",
  0x17, 4, 1, 1, 1, "t", "T",
  0x1C, 5, 1, 1, 1, "y", "Y",
  0x18, 6, 1, 1, 1, "u", "U",
  0x0C, 7, 1, 1, 1, "i", "I",
  0x12, 8, 1, 1, 1, "o", "O",
  0x13, 9, 1, 1, 1, "p", "P",
  0x2F,10, 1, 1, 1, "@", "`",
  
  0x04, 0, 2, 1, 1, "a", "A",
  0x16, 1, 2, 1, 1, "s", "S",
  0x07, 2, 2, 1, 1, "d", "D",
  0x09, 3, 2, 1, 1, "f", "F",
  0x0A, 4, 2, 1, 1, "g", "G",
  0x0B, 5, 2, 1, 1, "h", "H",
  0x0D, 6, 2, 1, 1, "j", "J",
  0x0E, 7, 2, 1, 1, "k", "K",
  0x0F, 8, 2, 1, 1, "l", "L",
  
  0x1D, 0, 3, 1, 1, "z", "Z",
  0x1B, 1, 3, 1, 1, "x", "X",
  0x06, 2, 3, 1, 1, "c", "C",
  0x19, 3, 3, 1, 1, "v", "V",
  0x05, 4, 3, 1, 1, "b", "B",
  0x11, 5, 3, 1, 1, "n", "N",
  0x10, 6, 3, 1, 1, "m", "M",
  0x36, 7, 3, 1, 1, ",", "<",
  0x37, 8, 3, 1, 1, ".", ">",
  
  0x2E, 0, 4, 1, 1, "^", "~",
  0x89, 1, 4, 1, 1, "\\", "|",
  0x33, 2, 4, 1, 1, ";", "+",
  0x34, 3, 4, 1, 1, ":", "*",
  0x30, 4, 4, 1, 1, "[", "{",
  0x32, 5, 4, 1, 1, "]", "}",
  0x38, 6, 4, 1, 1, "/", "?",
  0x87, 7, 4, 1, 1, "\\", "_",

  0x49, 8, 4, 1, 1, "Ins", "Ins",
  0x4C, 8, 5, 1, 1, "Del", "Del",

  0x2A, 9, 2, 2, 1, "BS", "BS",
  0x28, 9, 3, 2, 2, "Ent", "Ent",
  0x29, 0, 5, 2, 1, "Esc", "Esc",
  0x35, 2, 5, 2, 1, "Chin", "Chin",
  0x2B, 4, 5, 2, 1, "Tab", "Tab",
//  0x39, 4, 5, 2, 1, "Caps", "Caps",
  0xE1, 6, 5, 2, 1, "Shift", "Shift",
  0xE0, 0, 6, 2, 1, "Ctrl", "Ctrl",
  0xE2, 2, 6, 2, 1, "Alt", "Alt",
  0x2C, 4, 6, 4, 1, "Space", "Space",
  0x52, 9, 5, 1, 1, "^", "^",
  0x51, 9, 6, 1, 1, "v", "v",
  0x50, 8, 6, 1, 1, "<", "<",
  0x4F,10, 6, 1, 1, ">", ">"
};

static const uint8_t font_han[] =
{
  0x00, 0x00,
  0x00, 0x00,
  0x40, 0x40,
  0x24, 0x80,
  0x15, 0x00,
  0x04, 0x00,
  0x7f, 0xc0,
  0x04, 0x00,
  0x04, 0x00,
  0x04, 0x00,
  0xff, 0xe0,
  0x04, 0x00,
  0x04, 0x00,
  0x04, 0x00,
  0x00, 0x00,
  0x00, 0x00
};
static const uint8_t font_kaku[] =
{
  0x00, 0x00,
  0x00, 0x00,
  0x1f, 0x80,
  0x21, 0x00,
  0x02, 0x00,
  0x7f, 0xc0,
  0x44, 0x40,
  0x44, 0x40,
  0x7f, 0xc0,
  0x44, 0x40,
  0x44, 0x40,
  0x7f, 0xc0,
  0x40, 0x40,
  0x80, 0x40,
  0x00, 0x00,
  0x00, 0x00
};
static const uint8_t font_zen[] =
{
  0x00, 0x00,
  0x00, 0x00,
  0x04, 0x00,
  0x0a, 0x00,
  0x11, 0x00,
  0x20, 0x80,
  0x40, 0x40,
  0xbf, 0xa0,
  0x04, 0x00,
  0x04, 0x00,
  0x3F, 0x80,
  0x04, 0x00,
  0x04, 0x00,
  0x7f, 0xc0,
  0x00, 0x00,
  0x00, 0x00
};
static const uint8_t font_sura[] =
{
  0x00,
  0x00,
  0x10,
  0x10,
  0x10,
  0x20,
  0x20,
  0x20,
  0x40,
  0x40,
  0x40,
  0x80,
  0x80,
  0x80,
  0x00,
  0x00,
};

int key_alt;
int key_shft;
int key_ctrl;

void setup() {
  int i;
  
  pinMode(LED, OUTPUT);
  Serial.begin( 115200 );
  delay(100);
  Serial.println("ESP32_ILI9341_BLE_keyboard Start...");

  tft.begin();
  tft.setRotation(3);
  tft.fillScreen(ILI9341_BLACK);
  ts.begin();
  ts.setRotation(3);

  drawFont16(0, YMAX/2-20, "Wait for", ILI9341_WHITE, ILI9341_BLACK);
  drawFont16(20, YMAX/2, "Bluetooth connection !", ILI9341_WHITE, ILI9341_BLACK);
  
  tft.display();

  ble = BleHidKeyboard::getInstance();
  ble->set_led_pin(LED);
  delay( 200 );
  ble->init("ESP32_keyboard");
  while(!ble->is_connected())
    delay(200);  

  key_alt=0;
  key_shft=0;
  key_ctrl=0;

  show_key(-1, key_alt, key_shft, key_ctrl);

  ble->report_release(0xE1, 0); // Release Shift
}
 
void loop() {
  struct TS_Point p;
  int mx, my;
  int i;
  char str[32];
  int  key;
  int  num;
  uint8_t modif;

  istouched=ts.touched();
  
  if (!istouched || wastouched){
    wastouched=istouched;
    delay(100);
    return;
  }

  wastouched=istouched;

  p=ts.getPoint();

  p.x = map(p.x, TS_MINX, TS_MAXX, 0, XMAX);
  p.y = map(p.y, TS_MINY, TS_MAXY, 0, YMAX);
  mx=XMAX-1-p.x;
  my=YMAX-1-p.y;

  key=-1;
  for(i=0;i<MAX_KEY;i++){
    if (key_code[i].x*SIZEX<=mx && key_code[i].y*SIZEY<=my &&
        mx<(key_code[i].x+key_code[i].sx)*SIZEX && my<(key_code[i].y+key_code[i].sy)*SIZEY){
          num=i;
          key=key_code[num].code;
          break;
    }
  }

  if (key>=0){
    show_key(num, key_alt, key_shft, key_ctrl);
    // Shift
    if (key==0xE1 || key==0xE5){
      if (key_shft==0){
        key_shft=1;
        sprintf(str, "DOWN : %02X", key);
        Serial.println(str);
        modif=key_ctrl+key_shft*2;
        ble->report_press(key, modif);
      }else{
        key_shft=0;
        sprintf(str, "UP : %02X", key);
        Serial.println(str);
        modif=key_ctrl+key_shft*2;
        ble->report_release(key, modif);
      }
    
    // Ctrl
    }else if (key==0xE0 || key==0xE4){
      modif=key_ctrl+key_shft*2;
      if (key_ctrl==0){
        key_ctrl=1;
        sprintf(str, "DOWN : %02X", key);
        Serial.println(str);
        modif=key_ctrl+key_shft*2;
        ble->report_press(key, modif);
      }else{
        key_ctrl=0;
        sprintf(str, "UP : %02X", key);
        Serial.println(str);
        modif=key_ctrl+key_shft*2;
        ble->report_release(key, modif);
      }

    // Alt
    }else if (key==0xE2 || key==0xE6){
      modif=key_ctrl+key_shft*2+key_alt*4;
      if (key_alt==0){
        key_alt=1;
        sprintf(str, "DOWN : %02X", key);
        Serial.println(str);
        modif=key_ctrl+key_shft*2+key_alt*4;
        ble->report_press(key, modif);
      }else{
        key_alt=0;
        sprintf(str, "UP : %02X", key);
        Serial.println(str);
        modif=key_ctrl+key_shft*2+key_alt*4;
        ble->report_release(key, modif);
      }

    }else{
      modif=key_ctrl+key_shft*2+key_alt*4;
      sprintf(str, "DOWN : %02X", key);
      Serial.println(str);
      ble->report_press(key, modif);
      delay(10);
      sprintf(str, "UP : %02X", key);
      Serial.println(str);
      ble->report_release(key, modif);
    }
    show_key(-1, key_alt, key_shft, key_ctrl);
  }

  delay(10);
}

void show_key(int key_num, int alt, int shft, int ctrl)
{
  int i;
  uint16_t bk;
  
  tft.fillScreen(ILI9341_BLACK);

  for(i=0;i<MAX_KEY;i++){
    bk=ILI9341_BLACK;
    if (shft==1 && key_code[i].code==0xE1){
      tft.fillRect(key_code[i].x*SIZEX, key_code[i].y*SIZEY, key_code[i].sx*SIZEX, key_code[i].sy*SIZEY, ILI9341_RED);
      bk=ILI9341_RED;
    }
    if (ctrl==1 && key_code[i].code==0xE0){
      tft.fillRect(key_code[i].x*SIZEX, key_code[i].y*SIZEY, key_code[i].sx*SIZEX, key_code[i].sy*SIZEY, ILI9341_RED);
      bk=ILI9341_RED;
    }
    if (alt==1 && key_code[i].code==0xE2){
      tft.fillRect(key_code[i].x*SIZEX, key_code[i].y*SIZEY, key_code[i].sx*SIZEX, key_code[i].sy*SIZEY, ILI9341_RED);
      bk=ILI9341_RED;
    }
    if (i==key_num){
      tft.fillRect(key_code[i].x*SIZEX, key_code[i].y*SIZEY, key_code[i].sx*SIZEX, key_code[i].sy*SIZEY, ILI9341_YELLOW);
      bk=ILI9341_YELLOW;
    }
 
    tft.drawRect(key_code[i].x*SIZEX, key_code[i].y*SIZEY, key_code[i].sx*SIZEX, key_code[i].sy*SIZEY, ILI9341_WHITE);

    if (key_code[i].code==0x35){
      tft.drawBitmap(key_code[i].x*SIZEX+2, key_code[i].y*SIZEY+6, font_han, 12, 16, ILI9341_WHITE, bk);
      tft.drawBitmap(key_code[i].x*SIZEX+2+12, key_code[i].y*SIZEY+6, font_kaku, 12, 16, ILI9341_WHITE, bk);
      tft.drawBitmap(key_code[i].x*SIZEX+2+24, key_code[i].y*SIZEY+6, font_sura, 6, 16, ILI9341_WHITE, bk);
      tft.drawBitmap(key_code[i].x*SIZEX+2+30, key_code[i].y*SIZEY+6, font_zen, 12, 16, ILI9341_WHITE, bk);
      tft.drawBitmap(key_code[i].x*SIZEX+2+42, key_code[i].y*SIZEY+6, font_kaku, 12, 16, ILI9341_WHITE, bk);
    }else if (key_code[i].code==0x49 || key_code[i].code==0x4c){
      drawFont16(key_code[i].x*SIZEX+2, key_code[i].y*SIZEY+6, key_code[i].name, ILI9341_WHITE, bk);
    }else if (shft==1){
      drawFont16(key_code[i].x*SIZEX+6, key_code[i].y*SIZEY+6, key_code[i].shift, ILI9341_WHITE, bk);
    }else{
      drawFont16(key_code[i].x*SIZEX+6, key_code[i].y*SIZEY+6, key_code[i].name, ILI9341_WHITE, bk);
    }
  }
  tft.display();
}

void drawFont16(int x, int y, char *str, int col, int bkcol)
{
  int i;
  for(i=0; str[i]!='\0'; i++)
    tft.drawBitmap(x+i*9, y, &font16[(str[i]-' ')*16], 8, 16, col, bkcol);
}
